/*  Copyright (C) 2015 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/SegmenterRegionGrowingfunctions.h
  \brief Useful functions for Region Growing segmentation.
 */

#ifndef __TERRALIB_RP_INTERNAL_SEGMENTERREGIONGROWINGFUNCTIONS_H
#define __TERRALIB_RP_INTERNAL_SEGMENTERREGIONGROWINGFUNCTIONS_H

#include "Matrix.h"
#include "SegmenterIdsManager.h"
#include "SegmenterRegionGrowingMerger.h"
#include "SegmenterRegionGrowingSegment.h"
#include "SegmenterSegmentsBlock.h"

namespace te
{
  namespace rp
  {
    namespace rg
    {
      /*!
        \brief Definition for segments ids matrix type.
      */
      typedef Matrix< SegmenterSegmentsBlock::SegmentIdDataType >
      SegmentsIdsMatrixT;

      /*!
        \brief Iteration counter type.
      */
      typedef unsigned short int IterationCounterType;

      /*!
        \brief Mean Strategy feature type.
      */
      typedef float MeanFeatureType;
      
      /*!
        \brief Baatz Strategy feature type.
      */
      typedef float BaatzFeatureType;
      
      /*!
        \brief Export the segments IDs to a tif file.
        \param segmentsIds The output segment ids container.
        \param normto8bits If true, a 8 bits file will be generated.
        \param fileName The output tif file name.
      */
      void exportSegs2Tif( const SegmentsIdsMatrixT& segmentsIds,
                           bool normto8bits, const std::string& fileName );

      /*!
        \brief Returns the number of active segments.
        \param actSegsListHeadPtr A pointer the the active segments list head.
        \return Returns the number of active segments.
      */
      template< typename FeatureDataTypeT >
      unsigned long int getActiveSegmentsNumber( SegmenterRegionGrowingSegment< FeatureDataTypeT >* const actSegsListHeadPtr )
      {
        unsigned long int returnValue = 0;
      
        SegmenterRegionGrowingSegment< FeatureDataTypeT >* currSegPtr = actSegsListHeadPtr;
      
        while( currSegPtr ) {
          ++returnValue;
          currSegPtr = currSegPtr->m_nextActiveSegment;
        }

        return returnValue;
      }

      /*!
        \brief Merge closest segments.
        \param segmentsIds Segments ids container.
        \param disimilarityThreshold The maximum similarity value allowed when deciding when to merge two segments.
        \param maximumSegmentSize Only merge neighbor segments with target segments with sizes smaller than this value.
        \param segmenterIdsManager A segments ids manager to acquire unique segments ids.
        \param merger The merger instance to use.
        \param enablelocalMutualBestFitting If enabled, a merge only occurs between two segments if the minimum dissimilarity criteria is best fulfilled mutually.
        \param enableSameIterationMerges If enabled, a merged segment could be merged with another under the same iteration.
        \param minFoundDissimilarity The minimum dissimilarity value found.
        \param maxFoundDissimilarity The maximum dissimilarity value found.
        \param totalMergesNumber The total number of merges.
        \param globalMergeIterationsCounter A reference to a iteration number counter (this variable will be only incremented, never zeroed. It never must be reset. ).
        \param actSegsListHeadPtr A pointer the the active segments list head.
        \param minMergesIterationThreshold Stop itarating when the number of merges drops below this value (Range 0-100);
      */
      template< typename FeatureDataTypeT >
      void mergeSegments( SegmentsIdsMatrixT& segmentsIds,
        const DissimilarityTypeT disimilarityThreshold,
        const unsigned int maximumSegmentSize,
        SegmenterIdsManager& segmenterIdsManager,
        SegmenterRegionGrowingMerger< FeatureDataTypeT >& merger,
        const bool enablelocalMutualBestFitting,
        const bool enableSameIterationMerges,
        DissimilarityTypeT& minFoundDissimilarity,
        DissimilarityTypeT& maxFoundDissimilarity,
        unsigned long int& totalMergesNumber,
        IterationCounterType& globalMergeIterationsCounter,
        SegmenterRegionGrowingSegment< FeatureDataTypeT >** const actSegsListHeadPtrPtr,
        const unsigned short int minMergesIterationThreshold )
      {
        TERP_TRUE_OR_THROW( actSegsListHeadPtrPtr != 0, 
          "Invalid active segments list header pointer" );
        
        if( (*actSegsListHeadPtrPtr) == 0 ) return;
      
        minFoundDissimilarity = std::numeric_limits< DissimilarityTypeT >::max();
        maxFoundDissimilarity = DissimilarityTypeT( -1.0 * minFoundDissimilarity );
        totalMergesNumber = 0;
      
        // Globals
        
        unsigned long int iterationTotalInitalSegmentsNumber = 0;
        unsigned long int minRequiredIterationMerges = 0;
        unsigned long int iterationMergedSegmentsNumber = 0;
        unsigned int segmentsLine = 0;
        unsigned int segmentsLineBound = 0;
        unsigned int segmentCol = 0;
        unsigned int segmentColStart = 0;
        unsigned int segmentColBound = 0;
        SegmenterRegionGrowingSegment< FeatureDataTypeT >* neighborSegmentPtr = 0;
        SegmenterRegionGrowingSegment< FeatureDataTypeT >* minForwardDissimilaritySegmentPtr = 0;
        DissimilarityTypeT forwardDissimilarityValue = 0;
        DissimilarityTypeT minForwardDissimilarityValue = 0;
        SegmenterRegionGrowingSegment< FeatureDataTypeT >* minBackwardDissimilaritySegmentPtr = 0;
        DissimilarityTypeT backwardDissimilarityValue = 0;
        DissimilarityTypeT minBackwardDissimilarityValue = 0;
        SegmenterSegmentsBlock::SegmentIdDataType* segmentsIdsLinePtr = 0;
        SegmenterSegmentsBlock::SegmentIdDataType currentSegmentId = 0;
        std::list< SegmenterSegmentsBlock::SegmentIdDataType > freeSegmentIds;
        unsigned int neighborSegIdx = 0;
        SegmenterRegionGrowingSegment< FeatureDataTypeT >* currActSegPtr = 
          (*actSegsListHeadPtrPtr);
          
        // Merge data          
          
        // forwardDissimilarityAuxSegPtr points to the current auxiliar 
        // forward merge data, the other
        // merge data instance contains the previous merge prewiew.
        // this pointer alternates between  and
        // forwardDissimilarityAuxMergeDataPtr1 and
        // forwardDissimilarityAuxMergeDataPtr2          
          
        SegmenterRegionGrowingMergeData* forwardDissimilarityAuxSegPtr = 0;
          
        merger.createMergeData( &forwardDissimilarityAuxSegPtr );
        std::unique_ptr< SegmenterRegionGrowingMergeData >
          forwardDissimilarityAuxMergeDataPtr1( 
          forwardDissimilarityAuxSegPtr );
          
        merger.createMergeData( &forwardDissimilarityAuxSegPtr );
        std::unique_ptr< SegmenterRegionGrowingMergeData >
          forwardDissimilarityAuxMergeDataPtr2(
          forwardDissimilarityAuxSegPtr );
          
        merger.createMergeData( &forwardDissimilarityAuxSegPtr );
        std::unique_ptr< SegmenterRegionGrowingMergeData >
          backwardDissimilarityAuxSeg( forwardDissimilarityAuxSegPtr );

        forwardDissimilarityAuxSegPtr = forwardDissimilarityAuxMergeDataPtr1.get();
          
        // Main iterations loop

        do 
        {
          iterationMergedSegmentsNumber = 0;
          currActSegPtr = (*actSegsListHeadPtrPtr);
          iterationTotalInitalSegmentsNumber = 
            getActiveSegmentsNumber( *actSegsListHeadPtrPtr );
          minRequiredIterationMerges = (unsigned long int)
            ( ((double)iterationTotalInitalSegmentsNumber) * ((double)minMergesIterationThreshold)
            / 100.0 );          
        
          // Updating the merger state        
          merger.update( *actSegsListHeadPtrPtr );
        
          // iterating over each segment
          do 
          {
//             std::cout << std::endl
//               << "processing segment:" << currActSegPtr->m_id
//               << std:: endl;
            
            if( 
                ( 
                  currActSegPtr->m_size 
                  < 
                  maximumSegmentSize 
                )
                &&
                (
                  enableSameIterationMerges 
                  || 
                  ( 
                    currActSegPtr->m_mergetIteration 
                    < 
                    globalMergeIterationsCounter 
                  ) 
                )
              ) 
            {
              // finding the neighbor segment with minimum dissimilary value
              // related to the current sement 
            
              minForwardDissimilaritySegmentPtr = 0;
              minForwardDissimilarityValue = std::numeric_limits< DissimilarityTypeT >::max();
            
              for( neighborSegIdx = 0 ; neighborSegIdx < 
                currActSegPtr->m_neighborSegmentsSize ;
                ++neighborSegIdx ) 
              {
                neighborSegmentPtr = currActSegPtr->m_neighborSegments[ 
                  neighborSegIdx ];
                
                if( neighborSegmentPtr != 0 )
                {
                  forwardDissimilarityValue = merger.getDissimilarity( 
                    currActSegPtr, neighborSegmentPtr,
                    forwardDissimilarityAuxSegPtr );
                  
                  if( 
                      ( 
                        forwardDissimilarityValue <= disimilarityThreshold 
                      )
                      &&
                      ( 
                        forwardDissimilarityValue < minForwardDissimilarityValue 
                      ) 
                    ) 
                  {
                    minForwardDissimilarityValue = forwardDissimilarityValue;
                    minForwardDissimilaritySegmentPtr = neighborSegmentPtr;
                    
                    if( forwardDissimilarityAuxSegPtr == forwardDissimilarityAuxMergeDataPtr1.get() )
                    {
                      forwardDissimilarityAuxSegPtr = forwardDissimilarityAuxMergeDataPtr2.get();
                    }
                    else
                    {
                      forwardDissimilarityAuxSegPtr = forwardDissimilarityAuxMergeDataPtr1.get();
                    }
                  }
                }
              }
              
              // Checking if the neighbor segment aready meged within the current iteration
              // Checking the neighbor maximum size
              
              if( 
                  ( minForwardDissimilaritySegmentPtr )
                  &&
                  (
                    (
                      ( !enableSameIterationMerges )
                      &&
                      (
                        minForwardDissimilaritySegmentPtr->m_mergetIteration 
                        >=
                        globalMergeIterationsCounter 
                      )
                    )
                  )
                )
              {
                minForwardDissimilaritySegmentPtr = 0;
              }

              // does the neighbor wants to merge back ?

              if(
                  ( minForwardDissimilaritySegmentPtr )
                  &&
                  enablelocalMutualBestFitting 
                )
              {
                // Calculating all neighbor neighbor segments dissimilarity          

                minBackwardDissimilaritySegmentPtr = 0;
                backwardDissimilarityValue = 0;
                minBackwardDissimilarityValue = std::numeric_limits< DissimilarityTypeT >::max();

                for( neighborSegIdx = 0 ;
                     neighborSegIdx < minForwardDissimilaritySegmentPtr->m_neighborSegmentsSize ;
                     ++neighborSegIdx )
                {
                  if( minForwardDissimilaritySegmentPtr->m_neighborSegments[ neighborSegIdx ] ) 
                  {
                    backwardDissimilarityValue = merger.getDissimilarity( 
                      minForwardDissimilaritySegmentPtr, 
                      minForwardDissimilaritySegmentPtr->m_neighborSegments[ neighborSegIdx ],
                      backwardDissimilarityAuxSeg.get() );

                    if( backwardDissimilarityValue < minBackwardDissimilarityValue )
                    {
                      minBackwardDissimilarityValue = backwardDissimilarityValue;
                      minBackwardDissimilaritySegmentPtr = 
                        minForwardDissimilaritySegmentPtr->m_neighborSegments[ neighborSegIdx ];
                    }
                  }
                }

                if( minBackwardDissimilaritySegmentPtr != currActSegPtr ) {
                  minForwardDissimilaritySegmentPtr = 0;
                }
              }

              // if the min forward dissimilarity segment was found
              if( minForwardDissimilaritySegmentPtr )
              {
                if( minFoundDissimilarity > minForwardDissimilarityValue ) {
                  minFoundDissimilarity = minForwardDissimilarityValue;
                }
    
                if( maxFoundDissimilarity < minForwardDissimilarityValue ) {
                  maxFoundDissimilarity = minForwardDissimilarityValue;
                }             
              
                // If the maximum similary neighbor was found it will be merged
                // if the dissimilarity value is below the threshold
                // merging segment data
                
                if( forwardDissimilarityAuxSegPtr == forwardDissimilarityAuxMergeDataPtr1.get() )
                {
                  forwardDissimilarityAuxSegPtr = forwardDissimilarityAuxMergeDataPtr2.get();
                }
                else
                {
                  forwardDissimilarityAuxSegPtr = forwardDissimilarityAuxMergeDataPtr1.get();
                }

                merger.mergeFeatures( currActSegPtr, minForwardDissimilaritySegmentPtr,
                  forwardDissimilarityAuxSegPtr );
              
                currActSegPtr->m_mergetIteration = globalMergeIterationsCounter;
                
                currActSegPtr->removeNeighborSegmentPtr( minForwardDissimilaritySegmentPtr );
                
                // updating the max similarity segment neighborhood segments
                // with the current segment
              
                for( neighborSegIdx = 0 ;
                     neighborSegIdx < minForwardDissimilaritySegmentPtr->m_neighborSegmentsSize;
                     ++neighborSegIdx ) 
                {
                  if( 
                     ( 
                       minForwardDissimilaritySegmentPtr->m_neighborSegments[ 
                       neighborSegIdx ] != 0 
                     )
                     &&
                     ( 
                       minForwardDissimilaritySegmentPtr->m_neighborSegments[ 
                       neighborSegIdx ] != currActSegPtr ) 
                     ) 
                  {
                    // adding the max similarity neighborhood segments to the 
                    // current one, if it is not already there            
                  
                    currActSegPtr->addNeighborSegmentPtr( 
                      minForwardDissimilaritySegmentPtr->m_neighborSegments[ 
                      neighborSegIdx ] );
                    
                    // adding the current segment into the max similarity 
                    // neighborhood segments list, if it is not already there              
                  
                    minForwardDissimilaritySegmentPtr->m_neighborSegments[ 
                      neighborSegIdx ]->addNeighborSegmentPtr( currActSegPtr );
                    
                    // removing the merged segment reference from its neighbor
                    // list      
                    minForwardDissimilaritySegmentPtr->m_neighborSegments[ 
                      neighborSegIdx ]->removeNeighborSegmentPtr( 
                      minForwardDissimilaritySegmentPtr );
                  }
                }      
              
                // updating the segments Ids container matrix
              
                segmentsLineBound = minForwardDissimilaritySegmentPtr->m_yBound;
                segmentColStart = minForwardDissimilaritySegmentPtr->m_xStart;
                segmentColBound = minForwardDissimilaritySegmentPtr->m_xBound;          
                currentSegmentId = currActSegPtr->m_id;
                
                for( segmentsLine = minForwardDissimilaritySegmentPtr->m_yStart ; 
                     segmentsLine < segmentsLineBound ; ++segmentsLine ) 
                {
                  segmentsIdsLinePtr = segmentsIds[ segmentsLine ];
                
                  for( segmentCol = segmentColStart ; segmentCol < segmentColBound ; 
                    ++segmentCol ) 
                  {
                    if( 
                        segmentsIdsLinePtr[ segmentCol ] 
                        ==
                        minForwardDissimilaritySegmentPtr->m_id 
                      ) 
                    {
                      segmentsIdsLinePtr[ segmentCol ] = currentSegmentId;
                    }
                  }
                }
              
                // disabling the  merged segment
                // The merged segment id will be given back to ids manager
              
                if( minForwardDissimilaritySegmentPtr == (*actSegsListHeadPtrPtr) ) 
                {
                  (*actSegsListHeadPtrPtr) = (*actSegsListHeadPtrPtr)->m_nextActiveSegment;
                }
              
                if( minForwardDissimilaritySegmentPtr->m_prevActiveSegment ) 
                {
                  minForwardDissimilaritySegmentPtr->m_prevActiveSegment->m_nextActiveSegment =
                    minForwardDissimilaritySegmentPtr->m_nextActiveSegment;
                }
                
                if( minForwardDissimilaritySegmentPtr->m_nextActiveSegment ) 
                {
                  minForwardDissimilaritySegmentPtr->m_nextActiveSegment->m_prevActiveSegment =
                    minForwardDissimilaritySegmentPtr->m_prevActiveSegment;
                }
              
                minForwardDissimilaritySegmentPtr->disable();
              
                minForwardDissimilaritySegmentPtr->removeAllNeighborSegmentsPtrs();
                
                freeSegmentIds.push_back( minForwardDissimilaritySegmentPtr->m_id );
              
                ++iterationMergedSegmentsNumber;
              }
            }
          
            // going to the next segment
            currActSegPtr = currActSegPtr->m_nextActiveSegment;
          
          } while( currActSegPtr );
        
          // give back the free unused sement ids
        
          if( ! freeSegmentIds.empty() ) {
            segmenterIdsManager.addFreeIDs( freeSegmentIds );
            freeSegmentIds.clear();
          }
        
          totalMergesNumber += iterationMergedSegmentsNumber;
        
          ++globalMergeIterationsCounter;
          
//           std::cout << std::endl
//             << " disimilarityThreshold:" << disimilarityThreshold
//             << " maximumSegmentSize:" << maximumSegmentSize
//             << " iterationTotalInitalSegmentsNumber:" << iterationTotalInitalSegmentsNumber
//             << " globalMergeIterationsCounter:" << globalMergeIterationsCounter
//             << " iterationMergedSegmentsNumber:" << iterationMergedSegmentsNumber
//             << " minRequiredIterationMerges:" << minRequiredIterationMerges
//             << " totalMergesNumber:" << totalMergesNumber
//             << std::endl;
        } while( 
                 ( iterationMergedSegmentsNumber  > minRequiredIterationMerges )
                 &&
                 ( 
                   globalMergeIterationsCounter 
                   <
                   std::numeric_limits< IterationCounterType >::max() 
                 ) 
               );
      }
    } // end namespace rg
  } // end namespace rp
} // end namespace te

#endif  // __TERRALIB_RP_INTERNAL_SEGMENTERREGIONGROWINGFUNCTIONS_H
